@@ -25,8 +25,13 @@
 
 #include <config.h>
 
-#include <fcntl.h>
+#ifndef _WIN32
 #include <unistd.h>
+#else
+#include <winsock.h>
+#endif
+
+#include <fcntl.h>
 #include <errno.h>
 #include <string.h>
 #include <stdlib.h>
@@ -39,7 +44,11 @@
 #include "gaim.h"
 #include "proxy.h"
 
-#include "pixmaps/irc_icon.xpm"
+#ifdef _WIN32
+#include "win32dep.h"
+#endif
+
+#include "pixmaps/protocols/irc/irc_icon.xpm"
 
 #define IRC_BUF_LEN 4096
 #define PDIWORDS 32
@@ -47,6 +56,10 @@
 #define USEROPT_SERV      0
 #define USEROPT_PORT      1
 
+/* for win32 compatability */
+G_MODULE_IMPORT GSList *connections;
+
+
 struct dcc_chat
 {
 	struct gaim_connection *gc;	
@@ -57,6 +70,21 @@ struct dcc_chat
 	char nick[80];		
 };
 
+struct irc_file_transfer {
+		enum { IFT_SENDFILE_IN, IFT_SENDFILE_OUT } type;
+		struct file_transfer *xfer;
+		char *sn;
+		char *name;
+		int len;
+		int watcher;
+		int awatcher;
+		char ip[12];
+		int port;
+		int fd;
+		int cur;
+		struct gaim_connection *gc;
+};
+
 GSList *dcc_chat_list = NULL;
 
 struct irc_data {
@@ -78,6 +106,7 @@ struct irc_data {
 	gboolean in_whois;
 	gboolean in_list;
 	GString *liststr;
+	GSList *file_transfers;
 };
 
 struct dcc_chat *
@@ -101,15 +130,14 @@ find_dcc_chat (struct gaim_connection *gc, char *nick)
 	return NULL;
 }
 
-static char *irc_name()
-{
-	return "IRC";
-}
-
 static int irc_write(int fd, char *data, int len)
 {
 	debug_printf("IRC C: %s", data);
+#ifndef _WIN32
 	return write(fd, data, len);
+#else
+	return send(fd, data, len, 0);
+#endif
 }
 
 static struct conversation *irc_find_chat(struct gaim_connection *gc, char *name)
@@ -425,7 +453,11 @@ dcc_chat_in (gpointer data, gint source, GaimInputCondition condition)
 	int n = 0, l;
 	struct conversation *convo;
 	debug_printf("THIS IS TOO MUCH EFFORT\n");
+#ifndef _WIN32
 	n = read (chat->fd, buffer, IRC_BUF_LEN);
+#else
+	n = recv (chat->fd, buffer, IRC_BUF_LEN, 0);
+#endif
 	if (n > 0)
 	  {
 		  
@@ -450,6 +482,52 @@ dcc_chat_in (gpointer data, gint source, GaimInputCondition condition)
 	}
 }
 
+static void irc_file_transfer_do(struct gaim_connection *gc, struct irc_file_transfer *ift) {
+	/* Ok, we better be receiving some crap here boyeee */
+	if (transfer_in_do(ift->xfer, ift->fd, ift->name, ift->len)) {
+			gaim_input_remove(ift->watcher);
+			ift->watcher = 0;
+	}
+}
+
+
+void irc_read_dcc_ack (gpointer data, gint source, GaimInputCondition condition) {
+		/* Read ACK Here */
+
+}
+
+void dcc_send_callback (gpointer data, gint source, GaimInputCondition condition) {
+		struct irc_file_transfer *ift = data;
+		struct irc_data *id = (struct irc_data *)ift->gc->proto_data;
+		struct sockaddr_in addr;
+		int len = sizeof(addr);
+		
+		addr.sin_family = AF_INET;
+		addr.sin_port = htons(ift->port);
+		addr.sin_addr.s_addr = INADDR_ANY;
+
+		ift->fd = accept(ift->fd, (struct sockaddr *)&addr, &len);
+		if (!ift->fd) {
+				/* FIXME: Handle this gracefully XXX */
+				printf("Something bad happened here, bubba!\n");
+				return;
+		}
+		
+	/*	ift->awatcher = gaim_input_add(ift->fd, GAIM_INPUT_READ, irc_read_dcc_ack, ift); */
+		
+		if (transfer_out_do(ift->xfer, ift->fd, 0)) {
+				gaim_input_remove(ift->watcher);
+				ift->watcher = 0;
+		}
+}
+
+void dcc_recv_callback (gpointer data, gint source, GaimInputCondition condition) {
+		struct irc_file_transfer *ift = data;
+
+		ift->fd = source;
+		irc_file_transfer_do(ift->gc, ift);
+}
+
 void dcc_chat_callback (gpointer data, gint source, GaimInputCondition condition) {
 	struct dcc_chat *chat = data;
 	struct conversation *convo = new_conversation (chat->nick);
@@ -605,6 +683,19 @@ static gboolean mode_has_arg(struct gaim_connection *gc, char sign, char mode)
 	return FALSE;
 }
 
+static void irc_chan_mode(struct gaim_connection *gc, char *room, char sign, char mode, char *argstr, char *who)
+{
+	struct conversation *c = irc_find_chat(gc, room);
+	char buf[IRC_BUF_LEN];
+	char *nick = g_strndup(who, strchr(who, '!') - who);
+
+	g_snprintf(buf, sizeof(buf), "%s sets mode %c%c %s on %s", 
+		   nick, sign, mode, strlen(argstr) ? argstr : "",
+		   room);
+	g_free(nick);
+	write_to_conv(c, buf, WFLAG_SYSTEM, NULL, time((time_t)NULL), -1);
+}
+
 static void irc_user_mode(struct gaim_connection *gc, char *room, char sign, char mode, char *nick)
 {
 	struct conversation *c = irc_find_chat(gc, room);
@@ -664,6 +755,7 @@ static void handle_mode(struct gaim_connection *gc, char *word[], char *word_eol
 	char sign = *modes++;
 	int arg = 1;
 	char *argstr;
+	char *who = word[1];
 
 	if (!c)
 		return;
@@ -686,11 +778,49 @@ static void handle_mode(struct gaim_connection *gc, char *word[], char *word_eol
 				argstr = "";
 			if (strchr(id->nickmodes, *modes))
 				irc_user_mode(gc, chan, sign, *modes, argstr);
+			else if (strchr(who, '!'))				
+				irc_chan_mode(gc, chan, sign, *modes, argstr, who);
 		}
 		modes++;
 	}
 }
 
+static void handle_version(struct gaim_connection *gc, char *word[], char *word_eol[], int num)
+{
+	struct irc_data *id = gc->proto_data;
+	GString *str;
+
+	id->liststr = g_string_new("");
+
+	id->liststr = g_string_append(id->liststr, "<b>Version: </b>");
+	id->liststr = g_string_append(id->liststr, word_eol[4]);
+
+	str = decode_html(id->liststr->str);
+	g_show_info_text(gc, NULL, 2, str->str, NULL);
+	g_string_free(str, TRUE);
+	g_string_free(id->liststr, TRUE);
+	id->liststr = NULL;
+}
+
+static void handle_who(struct gaim_connection *gc, char *word[], char *word_eol[], int num)
+{
+	struct irc_data *id = gc->proto_data;
+	char buf[IRC_BUF_LEN];
+
+	if (!id->in_whois) {
+		id->in_whois = TRUE;
+		id->liststr = g_string_new("");
+	}
+
+	switch (num) {
+	case 352:
+		g_snprintf(buf, sizeof(buf), "<b>%s</b> (%s@%s): %s<br>", 
+			   word[8], word[5], word[6], word_eol[11]);
+		id->liststr = g_string_append(id->liststr, buf);
+		break;
+	}
+}
+
 /* Handle our whois stuff here.  You know what, I have a sore throat.  You know
  * what I think about that? I'm not too pleased with it.  Perhaps I should take
  * some medicine, or perhaps I should go to bed? Blah!! */
@@ -730,9 +860,18 @@ static void handle_whois(struct gaim_connection *gc, char *word[], char *word_eo
 		default:
 			break;
 	}
-	
+
 	if (word_eol[5][0] == ':')
 		id->liststr = g_string_append(id->liststr, word_eol[5] + 1);
+	/* Nicer idle time output, by jonas@birme.se */
+	else if (isdigit(word_eol[5][0])) {
+		time_t idle = atol(word_eol[5]);
+		time_t signon = atol(strchr(word_eol[5], ' '));
+				
+		g_snprintf(tmp, sizeof(tmp), 
+			   "%ld seconds [signon: %s]", (idle / 1000), ctime(&signon));
+		id->liststr = g_string_append(id->liststr, tmp);
+	}
 	else
 		id->liststr = g_string_append(id->liststr, word_eol[5]);
 }
@@ -752,6 +891,14 @@ static void handle_roomlist(struct gaim_connection *gc, char *word[], char *word
 	id->liststr = g_string_append(id->liststr, word_eol[4]);
 }
 
+static void irc_change_nick(void *a, char *b) {
+	struct gaim_connection *gc = a;
+	struct irc_data *id = gc->proto_data;
+	char buf[IRC_BUF_LEN];	
+	g_snprintf(buf, sizeof(buf), "NICK %s\r\n", b);
+	irc_write(id->fd, buf, strlen(buf));
+}
+
 static void process_numeric(struct gaim_connection *gc, char *word[], char *word_eol[])
 {
 	struct irc_data *id = gc->proto_data;
@@ -791,14 +938,16 @@ static void process_numeric(struct gaim_connection *gc, char *word[], char *word
 	case 312:
 	case 313:
 	case 317:
-	case 319:
+	case 319:		
 		handle_whois(gc, word, word_eol, n);
 		break;
 	case 322:
 		handle_roomlist(gc, word, word_eol);
 		break;
+	/* End of replies */
 	case 323:
 	case 318:
+	case 315:
 		if ((id->in_whois || id->in_list) && id->liststr) {
 			GString *str = decode_html(id->liststr->str);
 			g_show_info_text(gc, NULL, 2, str->str, NULL);
@@ -815,6 +964,12 @@ static void process_numeric(struct gaim_connection *gc, char *word[], char *word
 	case 332:
 		handle_topic(gc, text);
 		break;
+	case 351:
+		handle_version(gc, word, word_eol, n);
+		break;
+	case 352:
+		handle_who(gc, word, word_eol, n);
+		break;
 	case 353:
 		handle_names(gc, word[5], word_eol[6]);
 		break;
@@ -822,12 +977,24 @@ static void process_numeric(struct gaim_connection *gc, char *word[], char *word
 		irc_request_buddy_update(gc);
 		break;
 	case 401:
-		do_error_dialog(_("No such nick/channel"), _("IRC Error"));
+		do_error_dialog(_("There is no such nick or channel on this IRC channel."), NULL, GAIM_ERROR);
 		break;
 	case 402:
-		do_error_dialog(_("No such server"), _("IRC Error"));
+		do_error_dialog(_("There is no such IRC Server"), NULL, GAIM_ERROR);
 	case 431:
-		do_error_dialog(_("No nickname given"), _("IRC Error"));
+		do_error_dialog(_("No IRC nickname given"), NULL, GAIM_ERROR);
+		break;
+	case 433:
+		do_prompt_dialog(_("That nick is already in use.  Please enter a new nick"), gc->displayname, gc, irc_change_nick, NULL);
+		break;
+	default:
+		if (n > 400 && n < 502) {
+			char errmsg[IRC_BUF_LEN];
+			char *errmsg1 = strrchr(text, ':');
+			g_snprintf(errmsg, sizeof(errmsg), "IRC Error %d", n);
+			if (errmsg)
+				do_error_dialog(errmsg, errmsg1 ? errmsg1+1 : NULL, GAIM_ERROR);
+		}
 		break;
 	}
 }
@@ -925,6 +1092,8 @@ static void dcc_chat_init(gpointer obj, struct dcc_chat *data) {
 	memcpy(chat, data, sizeof(struct dcc_chat));  /* we have to make a new one
 						       * because the old one get's freed by
 						       * dcc_chat_cancel. */
+
+	printf("ONE MORE TIME: %s:%d\n", chat->ip_address, chat->port);
 	proxy_connect(chat->ip_address, chat->port, dcc_chat_callback, chat);
 }
 
@@ -971,12 +1140,34 @@ static void handle_ctcp(struct gaim_connection *gc, char *to, char *nick,
 		struct dcc_chat * dccchat = g_new0(struct dcc_chat, 1);
 		dccchat->gc = gc;	
 		g_snprintf(dccchat->ip_address, sizeof(dccchat->ip_address), chat_args[3]);	
+		printf("DCC CHAT DEBUG CRAP: %s\n", dccchat->ip_address);
 		dccchat->port=atoi(chat_args[4]);		
 		g_snprintf(dccchat->nick, sizeof(dccchat->nick), nick);	
 		g_snprintf(ask, sizeof(ask), _("%s has requested a DCC chat.  "
 					       "Would you like to establish the direct connection?"), nick);
 		do_ask_dialog(ask, dccchat, dcc_chat_init, dcc_chat_cancel);
 	}
+
+
+	if (!g_strncasecmp(msg, "DCC SEND", 8)) {
+			struct irc_file_transfer *ift = g_new0(struct irc_file_transfer, 1);
+			char **send_args = g_strsplit(msg, " ", 6);
+			send_args[5][strlen(send_args[5])-1] = 0;
+
+			ift->type = IFT_SENDFILE_IN;
+			ift->sn = g_strdup(nick);
+			ift->gc = gc;
+			g_snprintf(ift->ip, sizeof(ift->ip), send_args[3]);	
+			ift->port = atoi(send_args[4]);
+			ift->len = atoi(send_args[5]);
+			ift->name = g_strdup(send_args[2]);
+			ift->cur = 0;
+
+			id->file_transfers = g_slist_append(id->file_transfers, ift);
+
+			ift->xfer = transfer_in_add(gc, nick, ift->name, ift->len, 1, NULL);
+	}
+	
 	/* XXX should probably write_to_conv or something here */
 }
 
@@ -990,7 +1181,7 @@ static gboolean irc_parse(struct gaim_connection *gc, char *buf)
 	char *cmd;
 
 	/* Check for errors */
-
+	
 	if (*buf != ':') {
 		if (!strncmp(buf, "NOTICE ", 7))
 			buf += 7;
@@ -1068,9 +1259,9 @@ static gboolean irc_parse(struct gaim_connection *gc, char *buf)
 				return FALSE;
 			gc->buddy_chats = g_slist_remove(gc->buddy_chats, c);
 			c->gc = NULL;
-			g_snprintf(outbuf, sizeof(outbuf), _("You have been kicked from %s: %s"),
-					word[3], *word_eol[5] == ':' ? word_eol[5] + 1 : word_eol[5]);
-			do_error_dialog(outbuf, _("IRC Error"));
+			g_snprintf(outbuf, sizeof(outbuf), _("You have been kicked from %s by %s:"),
+				   word[3], nick);
+			do_error_dialog(outbuf, *word_eol[5] == ':' ? word_eol[5] + 1 : word_eol[5], GAIM_INFO);
 		} else {
 			char *reason = *word_eol[5] == ':' ? word_eol[5] + 1 : word_eol[5];
 			char *msg = g_strdup_printf(_("Kicked by %s: %s"), nick, reason);
@@ -1159,7 +1350,11 @@ static void irc_callback(gpointer data, gint source, GaimInputCondition conditio
 	gchar buf[1024];
 	gboolean off;
 
+#ifndef _WIN32
 	i = read(idata->fd, buf, 1024);
+#else
+	i = recv(idata->fd, buf, 1024, 0);
+#endif
 	if (i <= 0) {
 		hide_login_progress_error(gc, "Read error");
 		signoff(gc);
@@ -1182,7 +1377,7 @@ static void irc_callback(gpointer data, gint source, GaimInputCondition conditio
 		d = g_strndup(idata->rxqueue, len);
 		g_strchomp(d);
 		debug_printf("IRC S: %s\n", d);
-
+		
 		idata->rxlen -= len;
 		if (idata->rxlen) {
 			char *tmp = g_strdup(e + 1);
@@ -1210,7 +1405,11 @@ static void irc_login_callback(gpointer data, gint source, GaimInputCondition co
 	char buf[IRC_BUF_LEN];
 
 	if (!g_slist_find(connections, gc)) {
+#ifndef _WIN32
 		close(source);
+#else
+		closesocket(source);
+#endif
 		return;
 	}
 
@@ -1229,8 +1428,10 @@ static void irc_login_callback(gpointer data, gint source, GaimInputCondition co
 	hostname[sizeof(hostname) - 1] = 0;
 	if (!*hostname)
 		g_snprintf(hostname, sizeof(hostname), "localhost");
-	g_snprintf(buf, sizeof(buf), "USER %s %s %s :Gaim (%s)\r\n",
-		   g_get_user_name(), hostname, gc->user->proto_opt[USEROPT_SERV], WEBSITE);
+	g_snprintf(buf, sizeof(buf), "USER %s %s %s :%s\r\n",
+		   g_get_user_name(), hostname, 
+		   gc->user->proto_opt[USEROPT_SERV], 
+		   gc->user->alias && strlen(gc->user->alias) ? gc->user->alias : "gaim");
 	if (irc_write(idata->fd, buf, strlen(buf)) < 0) {
 		hide_login_progress(gc, "Write error");
 		signoff(gc);
@@ -1260,7 +1461,7 @@ static void irc_login(struct aim_user *user)
 	set_login_progress(gc, 2, buf);
 
 	idata->chantypes = g_strdup("#&!+");
-	idata->chanmodes = g_strdup("beI,k,l");
+	idata->chanmodes = g_strdup("beI,k,lnt");
 	idata->nickmodes = g_strdup("ohv");
 	idata->str = g_string_new("");
 
@@ -1278,16 +1479,38 @@ static void irc_login(struct aim_user *user)
 static void irc_close(struct gaim_connection *gc)
 {
 	struct irc_data *idata = (struct irc_data *)gc->proto_data;
+
 	gchar buf[IRC_BUF_LEN];
 
-	g_snprintf(buf, sizeof(buf), "QUIT :Download Gaim [%s]\r\n", WEBSITE);
+	if (idata->str->len > 0) {
+		g_snprintf(buf, sizeof(buf), "QUIT :%s\r\n", idata->str->str);
+	} else {
+		g_snprintf(buf, sizeof(buf), 
+			   "QUIT :Download Gaim [%s]\r\n", WEBSITE);
+	}
 	irc_write(idata->fd, buf, strlen(buf));
 
 	if (idata->rxqueue)
 		g_free(idata->rxqueue);
+
 	idata->rxqueue = NULL;
 	idata->rxlen = 0;
 
+	/* Kill any existing transfers */
+	while (idata->file_transfers) {
+			struct irc_file_transfer *ift = (struct irc_file_transfer *)idata->file_transfers->data;
+
+			g_free(ift->sn);
+			g_free(ift->name);
+			gaim_input_remove(ift->watcher);
+
+			close(ift->fd);
+
+			idata->file_transfers = idata->file_transfers->next;
+	}
+	idata->file_transfers = NULL;
+
+
 	g_free(idata->chantypes);
 	g_free(idata->chanmodes);
 	g_free(idata->nickmodes);
@@ -1302,30 +1525,14 @@ static void irc_close(struct gaim_connection *gc)
 	if (gc->inpa)
 		gaim_input_remove(gc->inpa);
 
+#ifndef _WIN32
 	close(idata->fd);
+#else
+	closesocket(idata->fd);
+#endif
 	g_free(gc->proto_data);
 }
 
-static GList *irc_user_opts()
-{
-	GList *m = NULL;
-	struct proto_user_opt *puo;
-
-	puo = g_new0(struct proto_user_opt, 1);
-	puo->label = "Server:";
-	puo->def = "irc.mozilla.org";
-	puo->pos = USEROPT_SERV;
-	m = g_list_append(m, puo);
-
-	puo = g_new0(struct proto_user_opt, 1);
-	puo->label = "Port:";
-	puo->def = "6667";
-	puo->pos = USEROPT_PORT;
-	m = g_list_append(m, puo);
-
-	return m;
-}
-
 static void set_mode_3(struct gaim_connection *gc, char *who, int sign, int mode,
 			int start, int end, char *word[])
 {
@@ -1437,6 +1644,17 @@ static void set_mode(struct gaim_connection *gc, char *who, int sign, int mode,
 	}
 }
 
+static void set_chan_mode(struct gaim_connection *gc, char *chan, char *mode_str)
+{
+	struct irc_data *id = gc->proto_data;
+	char buf[IRC_BUF_LEN];
+
+	if ((mode_str[0] == '-') || (mode_str[0] == '+')) {
+		g_snprintf(buf, sizeof(buf), "MODE %s %s\r\n", chan, mode_str);
+		irc_write(id->fd, buf, strlen(buf));
+	}
+}
+
 static int handle_command(struct gaim_connection *gc, char *who, char *what)
 {
 	char buf[IRC_BUF_LEN];
@@ -1496,11 +1714,14 @@ static int handle_command(struct gaim_connection *gc, char *who, char *what)
 		irc_write(id->fd, buf, strlen(buf));
 	} else if (!g_strcasecmp(pdibuf, "TOPIC")) {
 		if (!*word_eol[2]) {
-			g_free(what);
-			return -EINVAL;
+			struct conversation *c;
+			c = irc_find_chat(gc, who);
+			g_snprintf(buf, sizeof(buf), _("Topic for %s is %s"), who, c->topic ? c->topic : "(no topic set)");
+			write_to_conv(c, buf, WFLAG_SYSTEM | WFLAG_NOLOG, NULL, time(NULL), -1);
+		} else {
+			g_snprintf(buf, sizeof(buf), "TOPIC %s :%s\r\n", who, word_eol[2]);
+			irc_write(id->fd, buf, strlen(buf));
 		}
-		g_snprintf(buf, sizeof(buf), "TOPIC %s :%s\r\n", who, word_eol[2]);
-		irc_write(id->fd, buf, strlen(buf));
 	} else if (!g_strcasecmp(pdibuf, "NICK")) {
 		if (!*word_eol[2]) {
 			g_free(what);
@@ -1516,6 +1737,9 @@ static int handle_command(struct gaim_connection *gc, char *who, char *what)
 		set_mode(gc, who, '+', 'v', word);
 	} else if (!g_strcasecmp(pdibuf, "DEVOICE")) {
 		set_mode(gc, who, '-', 'v', word);
+	} else if (!g_strcasecmp(pdibuf, "MODE")) {
+		char *chan = who;
+		set_chan_mode(gc, chan, word_eol[2]);
 	} else if (!g_strcasecmp(pdibuf, "QUOTE")) {
 		if (!*word_eol[2]) {
 			g_free(what);
@@ -1579,7 +1803,7 @@ static int handle_command(struct gaim_connection *gc, char *who, char *what)
 			gc->buddy_chats = g_slist_remove(gc->buddy_chats, c);
 			c->gc = NULL;
 			g_snprintf(buf, sizeof(buf), _("You have left %s"), chan);
-			do_error_dialog(buf, _("IRC Part"));
+			do_error_dialog(buf, NULL, GAIM_INFO);
 		}
 	} else if (!g_strcasecmp(pdibuf, "WHOIS")) {
 		g_snprintf(buf, sizeof(buf), "WHOIS %s\r\n", word_eol[2]);
@@ -1587,6 +1811,16 @@ static int handle_command(struct gaim_connection *gc, char *who, char *what)
 	} else if (!g_strcasecmp(pdibuf, "LIST")) {
 		g_snprintf(buf, sizeof(buf), "LIST\r\n");
 		irc_write(id->fd, buf, strlen(buf));
+	} else if (!g_strcasecmp(pdibuf, "QUIT")) {
+		char *reason = word_eol[2];
+		id->str = g_string_insert(id->str, 0, reason);
+		do_quit();
+	} else if (!g_strcasecmp(pdibuf, "VERSION")) {
+		g_snprintf(buf, sizeof(buf), "VERSION\r\n");
+		irc_write(id->fd, buf, strlen(buf));
+	} else if (!g_strcasecmp(pdibuf, "W")) {
+		g_snprintf(buf, sizeof(buf), "WHO *\r\n");
+		irc_write(id->fd, buf, strlen(buf));
 	} else if (!g_strcasecmp(pdibuf, "HELP")) {
 		struct conversation *c = NULL;
 		if (is_channel(gc, who)) {
@@ -1602,7 +1836,8 @@ static int handle_command(struct gaim_connection *gc, char *who, char *what)
 				 "WHOIS INVITE NICK LIST<BR>"
 				 "JOIN PART TOPIC KICK<BR>"
 				 "OP DEOP VOICE DEVOICE<BR>"
-				 "ME MSG QUOTE SAY</B>",
+				 "ME MSG QUOTE SAY QUIT<BR>"
+			         "MODE VERSION W</B>",
 				 WFLAG_NOLOG, NULL, time(NULL), -1);
 	} else {
 		struct conversation *c = NULL;
@@ -1643,7 +1878,7 @@ static int send_msg(struct gaim_connection *gc, char *who, char *what)
 		return handle_command(gc, who, what);
 }
 
-static void irc_chat_invite(struct gaim_connection *gc, int idn, char *message, char *name) {
+static void irc_chat_invite(struct gaim_connection *gc, int idn, const char *message, const char *name) {
 	char buf[IRC_BUF_LEN]; 
 	struct irc_data *id = gc->proto_data;
 	struct conversation *c = irc_find_chat_by_id(gc, idn);
@@ -1659,7 +1894,7 @@ static int irc_send_im(struct gaim_connection *gc, char *who, char *what, int le
 }
 
 /* IRC doesn't have a buddy list, but we can still figure out who's online with ISON */
-static void irc_add_buddy(struct gaim_connection *gc, char *who) {}
+static void irc_add_buddy(struct gaim_connection *gc, const char *who) {}
 static void irc_remove_buddy(struct gaim_connection *gc, char *who, char *group) {}
 
 static GList *irc_chat_info(struct gaim_connection *gc)
@@ -1796,6 +2031,132 @@ static void dcc_chat_connected(gpointer data, gint source, GdkInputCondition con
 	dcc_chat_list = g_slist_append (dcc_chat_list, chat);
 }
 
+static void irc_ask_send_file(struct gaim_connection *gc, char *destsn) {
+		struct irc_data *id = (struct irc_data *)gc->proto_data;
+		struct irc_file_transfer *ift = g_new0(struct irc_file_transfer, 1);
+		char *localip = (char *)malloc(12);
+
+		if (getlocalip(localip) == -1) {
+				free(localip);
+				return;
+		} 
+
+		ift->type = IFT_SENDFILE_OUT;
+		ift->sn = g_strdup(destsn);
+		ift->gc = gc;
+		snprintf(ift->ip, sizeof(ift->ip), "%s", localip);
+		id->file_transfers = g_slist_append(id->file_transfers, ift);
+
+		ift->xfer = transfer_out_add(gc, ift->sn);
+}
+
+static struct irc_file_transfer *find_ift_by_xfer(struct gaim_connection *gc, 
+				struct file_transfer *xfer) {
+		
+		GSList *g = ((struct irc_data *)gc->proto_data)->file_transfers;
+		struct irc_file_transfer *f = NULL;
+
+		while (g) {
+				f = (struct irc_file_transfer *)g->data;
+				if (f->xfer == xfer)
+						break;
+				g = g->next;
+				f = NULL;
+		}
+
+		return f;
+}
+
+static void irc_file_transfer_data_chunk(struct gaim_connection *gc, struct file_transfer *xfer, const char *data, int len) {
+	struct irc_file_transfer *ift = find_ift_by_xfer(gc, xfer);
+	guint32 pos;
+	
+	ift->cur += len;
+	pos = htonl(ift->cur);
+	write(ift->fd, (char *)&pos, 4);
+
+	// FIXME: You should check to verify that they received the data when
+	// you are sending a file ...
+}
+
+static void irc_file_transfer_cancel (struct gaim_connection *gc, struct file_transfer *xfer) {
+	struct irc_data *id = (struct irc_data *)gc->proto_data;
+	struct irc_file_transfer *ift = find_ift_by_xfer(gc, xfer);
+
+	printf("Our shit got canceled, yo!\n");
+
+	/* Remove the FT from our list of transfers */
+	id->file_transfers = g_slist_remove(id->file_transfers, ift);
+
+	gaim_input_remove(ift->watcher);
+
+	/* Close our FT because we're done */
+	close(ift->fd);
+
+	g_free(ift->sn);
+	g_free(ift->name);
+
+	g_free(ift);
+}
+
+static void irc_file_transfer_done(struct gaim_connection *gc, struct file_transfer *xfer) {
+	struct irc_data *id = (struct irc_data *)gc->proto_data;
+	struct irc_file_transfer *ift = find_ift_by_xfer(gc, xfer);
+
+
+	printf("Our shit be done, yo.\n");
+	
+	/* Remove the FT from our list of transfers */
+	id->file_transfers = g_slist_remove(id->file_transfers, ift);
+
+	gaim_input_remove(ift->watcher);
+
+	/* Close our FT because we're done */
+	close(ift->fd);
+
+	g_free(ift->sn);
+	g_free(ift->name);
+
+	g_free(ift);
+}
+
+static void irc_file_transfer_out (struct gaim_connection *gc, struct file_transfer *xfer, const char *name, int totfiles, int totsize) {
+	struct irc_data *id = (struct irc_data *)gc->proto_data;
+	struct irc_file_transfer *ift = find_ift_by_xfer(gc, xfer);
+	struct sockaddr_in addr;
+	char buf[IRC_BUF_LEN];
+	int len;
+	
+	
+	ift->fd = socket (AF_INET, SOCK_STREAM, 0);
+	addr.sin_family = AF_INET;
+	addr.sin_port = 0;
+	addr.sin_addr.s_addr = INADDR_ANY;
+	bind (ift->fd, (struct sockaddr *) &addr, sizeof(addr));
+	listen(ift->fd, 1);
+
+	len = sizeof(addr);
+	getsockname (ift->fd, (struct sockaddr *) &addr, &len);
+
+	ift->port = ntohs(addr.sin_port);
+
+	ift->watcher = gaim_input_add (ift->fd, GAIM_INPUT_READ, dcc_send_callback, ift);
+	printf("watcher is %d\n", ift->watcher);
+
+	snprintf(buf, sizeof(buf), "\001DCC SEND %s %s %d %d\001\n", name, ift->ip, ift->port, totsize);
+	printf("Trying: %s\n", buf);
+	irc_send_im (gc, ift->sn, buf, -1, 0);
+}
+
+static void irc_file_transfer_in(struct gaim_connection *gc,
+				struct file_transfer *xfer, int offset) {
+
+	struct irc_file_transfer *ift = find_ift_by_xfer(gc, xfer);
+
+	ift->xfer = xfer;
+	proxy_connect(ift->ip, ift->port, dcc_recv_callback, ift);
+}
+
 static void irc_start_chat(struct gaim_connection *gc, char *who) {
 	struct dcc_chat *chat;
 	int len;
@@ -1823,7 +2184,7 @@ static void irc_start_chat(struct gaim_connection *gc, char *who) {
 	chat->inpa =
 		gaim_input_add (chat->fd, GAIM_INPUT_READ, dcc_chat_connected,
 			       chat);
-	snprintf (buf, sizeof buf, "\001DCC CHAT chat %s %d\001\n",
+	g_snprintf (buf, sizeof buf, "\001DCC CHAT chat %s %d\001\n",
 		  chat->ip_address, chat->port);
 	irc_send_im (gc, who, buf, -1, 0);
 }
@@ -1852,22 +2213,30 @@ static GList *irc_buddy_menu(struct gaim_connection *gc, char *who)
 	pbm->callback = irc_get_info;
 	pbm->gc = gc;
 	m = g_list_append(m, pbm);
+
 	pbm = g_new0(struct proto_buddy_menu, 1);
 	pbm->label = _("DCC Chat");
 	pbm->callback = irc_start_chat;
 	pbm->gc = gc;
 	m = g_list_append(m, pbm);
+/*
+	pbm = g_new0(struct proto_buddy_menu, 1);
+	pbm->label = _("DCC Send");
+	pbm->callback = irc_ask_send_file;
+	pbm->gc = gc;
+	m = g_list_append(m, pbm);
+*/
 	return m;
 }
 
 static struct prpl *my_protocol = NULL;
 
-void irc_init(struct prpl *ret)
+G_MODULE_EXPORT void irc_init(struct prpl *ret)
 {
+	struct proto_user_opt *puo;
 	ret->protocol = PROTO_IRC;
 	ret->options = OPT_PROTO_CHAT_TOPIC | OPT_PROTO_NO_PASSWORD;
-	ret->name = irc_name;
-	ret->user_opts = irc_user_opts;
+	ret->name = g_strdup("IRC");
 	ret->list_icon = irc_list_icon;
 	ret->login = irc_login;
 	ret->close = irc_close;
@@ -1884,32 +2253,32 @@ void irc_init(struct prpl *ret)
 	ret->buddy_menu = irc_buddy_menu;
 	ret->chat_invite = irc_chat_invite;
 	ret->convo_closed = irc_convo_closed;
+	ret->file_transfer_out = irc_file_transfer_out; 
+	ret->file_transfer_in = irc_file_transfer_in;
+	ret->file_transfer_data_chunk = irc_file_transfer_data_chunk;
+	ret->file_transfer_done = irc_file_transfer_done;
+	ret->file_transfer_cancel =irc_file_transfer_cancel;
+
+	puo = g_new0(struct proto_user_opt, 1);
+	puo->label = g_strdup("Server:");
+	puo->def = g_strdup("irc.openprojects.net");
+	puo->pos = USEROPT_SERV;
+	ret->user_opts = g_list_append(ret->user_opts, puo);
+
+	puo = g_new0(struct proto_user_opt, 1);
+	puo->label = g_strdup("Port:");
+	puo->def = g_strdup("6667");
+	puo->pos = USEROPT_PORT;
+	ret->user_opts = g_list_append(ret->user_opts, puo);
+
 	my_protocol = ret;
 }
 
 #ifndef STATIC
-
-char *gaim_plugin_init(GModule *handle)
+G_MODULE_EXPORT void gaim_prpl_init(struct prpl* prpl)
 {
-	load_protocol(irc_init, sizeof(struct prpl));
-	return NULL;
-}
-
-void gaim_plugin_remove()
-{
-	struct prpl *p = find_prpl(PROTO_IRC);
-	if (p == my_protocol)
-		unload_protocol(p);
-}
-
-char *name()
-{
-	return "IRC";
-}
-
-char *description()
-{
-	return PRPL_DESC("IRC");
+	irc_init(prpl);
+	prpl->plug->desc.api_version = PLUGIN_API_VERSION;
 }
 
 #endif
